from dataclasses import asdict
from typing import Any, Mapping

import numpy as np
from jinja2 import Environment, FileSystemLoader, Template

from .backend import Request
from .base import HTML, Input, Output, Prompt


class SimplePrompt(Prompt[str, str]):
    """
    A simple prompt that echos the string it is passed and returns the
    output of the LLM as a string.
    """

    def prompt(self, inp: str) -> Request:
        return Request(inp)

    def parse(self, out: str, inp: str) -> str:
        return out


class TemplatePrompt(Prompt[Mapping[str, Any], Output]):
    """
    A prompt that uses Jinja to define a prompt based on a static template.
    Set `template_file` to the Jinja template file.
    """

    IN = Mapping[str, Any]
    template = None
    template_file = ""
    prompt_template = ""
    stop_template = ""

    def render_prompt_html(self, inp: Mapping[str, Any], prompt: str) -> HTML:
        n = {}
        if not isinstance(inp, dict):
            inp = asdict(inp)

        for k, v in inp.items():
            if isinstance(v, str):
                n[k] = f"<div style='color:red'>{v}</div>"
            else:
                n[k] = v
        return HTML(self.prompt(n).prompt.replace("\n", "<br>"))

    def parse(self, result: str, inp: Mapping[str, Any]) -> Output:
        return str(result)  # type: ignore

    def prompt(self, kwargs: Mapping[str, Any]) -> Request:
        if self.template_file:
            tmp = Environment(loader=FileSystemLoader(".")).get_template(
                name=self.template_file
            )
        elif self.template:
            tmp = self.template  # type: ignore
        else:
            tmp = Template(self.prompt_template)
        if isinstance(kwargs, dict):
            x = tmp.render(**kwargs)
        else:
            x = tmp.render(**asdict(kwargs))

        if self.stop_template:
            stop = [Template(self.stop_template).render(**kwargs)]
        else:
            stop = None
        return Request(x, stop)


class EmbeddingPrompt(Prompt[Input, Output]):
    """
    A prompt that replaces parse with `find` that takes an embedding.
    """

    def parse(self, response: str, inp: Input) -> Output:
        return self.find(response, inp)

    def find(self, response: np.ndarray, inp: Input) -> Output:
        """
        Convert from the embedding response of the function
        to the output type.
        """
        raise NotImplementedError



from .backend import (
    Backend,
    Bash,
    Google,
    HuggingFaceEmbed,
    Id,
    Manifest,
    Mock,
    OpenAI,
    OpenAIEmbed,
    OpenAIStream,
    Python,
    Request,
    set_minichain_log,
    start_chain,
)
from .base import prompt, type_to_prompt
from .gradio import show



import os
import subprocess
import time
from dataclasses import dataclass
from types import TracebackType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Tuple

from eliot import start_action, to_file

if TYPE_CHECKING:
    import manifest


@dataclass
class Request:
    prompt: str
    stop: Optional[Sequence[str]] = None


class Backend:
    needs_request = True

    @property
    def description(self):
        return ""

    def run(self, request: Request) -> str:
        raise NotImplementedError

    async def arun(self, request: Request) -> str:
        return self.run(request)


class Id(Backend):
    def run(self, request: Request) -> str:
        return request.prompt


class Mock(Backend):
    def __init__(self, answers: List[str] = []):
        self.i = -1
        self.answers = answers

    def run(self, request: Request) -> str:
        self.i += 1
        return self.answers[self.i % len(self.answers)]

    def run_stream(self, request: Request) -> str:
        self.i += 1
        result = self.answers[self.i % len(self.answers)]
        for c in result:
            yield c
            time.sleep(10)

    def __repr__(self) -> str:
        return f"Mocked Backend {self.answers}"


class Google(Backend):
    def __init__(self) -> None:
        pass

    def run(self, request: Request) -> str:
        from serpapi import GoogleSearch

        serpapi_key = os.environ.get("SERP_KEY")
        assert (
            serpapi_key
        ), "Need a SERP_KEY. Get one here https://serpapi.com/users/welcome"
        self.serpapi_key = serpapi_key

        params = {
            "api_key": self.serpapi_key,
            "engine": "google",
            "q": request.prompt,
            "google_domain": "google.com",
            "gl": "us",
            "hl": "en",
        }

        search = GoogleSearch(params)
        res = search.get_dict()

        if "answer_box" in res.keys() and "answer" in res["answer_box"].keys():
            toret = res["answer_box"]["answer"]
        elif "answer_box" in res.keys() and "snippet" in res["answer_box"].keys():
            toret = res["answer_box"]["snippet"]
        elif (
            "answer_box" in res.keys()
            and "snippet_highlighted_words" in res["answer_box"].keys()
        ):
            toret = res["answer_box"]["snippet_highlighted_words"][0]
        elif "snippet" in res["organic_results"][0].keys():
            toret = res["organic_results"][0]["snippet"]
        else:
            toret = ""
        return str(toret)

    def __repr__(self) -> str:
        return "Google Search Backend"


class Python(Backend):
    """Executes Python commands and returns the output."""

    def run(self, request: Request) -> str:
        """Run commands and return final output."""
        from contextlib import redirect_stdout
        from io import StringIO

        p = request.prompt.strip()
        if p.startswith("```"):
            p = "\n".join(p.strip().split("\n")[1:-1])

        f = StringIO()
        with redirect_stdout(f):
            exec(p)
        s = f.getvalue()
        return s

    def __repr__(self) -> str:
        return "Python-Backend"


class Bash(Backend):
    """Executes bash commands and returns the output."""

    def __init__(self, strip_newlines: bool = False, return_err_output: bool = False):
        """Initialize with stripping newlines."""
        self.strip_newlines = strip_newlines
        self.return_err_output = return_err_output

    def run(self, request: Request) -> str:
        """Run commands and return final output."""
        try:
            output = subprocess.run(
                request.prompt,
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            ).stdout.decode()
        except subprocess.CalledProcessError as error:
            if self.return_err_output:
                return str(error.stdout.decode())
            return str(error)
        if self.strip_newlines:
            output = output.strip()
        return output

    def __repr__(self) -> str:
        return "Bash-Backend"


class OpenAIBase(Backend):
    def __init__(
        self,
        model: str = "text-davinci-003",
        max_tokens: int = 256,
        temperature: float = 0.0,
    ) -> None:
        self.model = model
        self.options = dict(
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
        )

    def __repr__(self) -> str:
        return f"OpenAI Backend {self.options}"


class OpenAIStream:
    def __init__(self, answers: List[str] = []):
        pass

    def run_stream(self, prompt):
        import openai

        self.api_key = os.environ.get("OPENAI_API_KEY")
        assert (
            self.api_key
        ), "Need an OPENAI_API_KEY. Get one here https://openai.com/api/"
        openai.api_key = self.api_key

        for chunk in openai.ChatCompletion.create(
            model="gpt-4",
            messages=[{"role": "user", "content": prompt}],
            stream=True,
        ):
            content = chunk["choices"][0].get("delta", {}).get("content")
            if content is not None:
                yield content

    def __repr__(self) -> str:
        return "OpenAI Stream Backend"


class OpenAI(OpenAIBase):
    def run(self, request: Request) -> str:
        import manifest

        manifest = manifest.Manifest(
            client_name="openai",
            max_tokens=self.options["max_tokens"],
            cache_name="sqlite",
            cache_connection=f"{MinichainContext.name}",
        )

        ans = manifest.run(
            # openai.Completion.create(
            # **self.options,
            # kstop=request.stop,
            # prompt=
            request.prompt,
            stop_sequences=request.stop,
        )
        return str(ans)

    async def arun(self, request: Request) -> str:
        import async_openai

        self.api_key = os.environ.get("OPENAI_API_KEY")
        async_openai.OpenAI.configure(
            api_key=self.api_key,
            debug_enabled=False,
        )
        ans = await async_openai.OpenAI.Completions.async_create(
            **self.options,
            stop=request.stop,
            prompt=request.prompt,
        )
        return str(ans.choices[0].text)


class OpenAIEmbed(OpenAIBase):
    def __init__(self, model: str = "text-embedding-ada-002", **kwargs: Any) -> None:
        super().__init__(model, **kwargs)

    def run(self, request: Request) -> str:
        import openai

        self.api_key = os.environ.get("OPENAI_API_KEY")
        assert (
            self.api_key
        ), "Need an OPENAI_API_KEY. Get one here https://openai.com/api/"
        openai.api_key = self.api_key

        ans = openai.Embedding.create(
            engine=self.model,
            input=request.prompt,
        )
        return ans["data"][0]["embedding"]  # type: ignore


class HuggingFaceBase(Backend):
    def __init__(self, model: str = "gpt2") -> None:
        self.model = model


class HuggingFace(HuggingFaceBase):
    def run(self, request: Request) -> str:

        from huggingface_hub.inference_api import InferenceApi

        self.api_key = os.environ.get("HF_KEY")
        assert self.api_key, "Need an HF_KEY. Get one here https://huggingface.co/"

        self.client = InferenceApi(
            token=self.api_key, repo_id=self.model, task="text-generation"
        )
        response = self.client(inputs=request.prompt)
        return response  # type: ignore


class HuggingFaceEmbed(HuggingFaceBase):
    def run(self, request: Request) -> str:

        from huggingface_hub.inference_api import InferenceApi

        self.api_key = os.environ.get("HF_KEY")
        assert self.api_key, "Need an HF_KEY. Get one here https://huggingface.co/"

        self.client = InferenceApi(
            token=self.api_key, repo_id=self.model, task="feature-extraction"
        )
        response = self.client(inputs=request.prompt)
        return response  # type: ignore


class Manifest(Backend):
    def __init__(self, client: "manifest.Manifest") -> None:
        "Client from [Manifest-ML](https://github.com/HazyResearch/manifest)."
        self.client = client

    def run(self, request: Request) -> str:
        try:
            import manifest
        except ImportError:
            raise ImportError("`pip install manifest-ml` to use the Manifest Backend.")
        assert isinstance(
            self.client, manifest.Manifest
        ), "Client must be a `manifest.Manifest` instance."

        return self.client.run(request.prompt)  # type: ignore


class MinichainContext:
    id_: int = 0
    prompt_store: Dict[Tuple[int, int], Tuple[Any, Request, str, Any]] = {}
    prompt_count: Dict[int, int] = {}
    name: str = ""


def set_minichain_log(name: str) -> None:
    to_file(open(f"{name}.log", "w"))


class MiniChain:
    """
    MiniChain session object with backends. Make backend by calling
    `minichain.OpenAI()` with args for `OpenAI` class.
    """

    def __init__(self, name: str):
        to_file(open(f"{name}.log", "w"))
        self.name = name

    def __enter__(self) -> "MiniChain":
        MinichainContext.prompt_store = {}
        MinichainContext.prompt_count = {}
        MinichainContext.name = self.name
        self.action = start_action(action_type=self.name)
        return self

    def __exit__(
        self,
        type: type,
        exception: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.action.finish()
        self.prompt_store = dict(MinichainContext.prompt_store)
        MinichainContext.prompt_store = {}
        MinichainContext.prompt_count = {}
        MinichainContext.name = ""


def start_chain(name: str) -> MiniChain:
    """
    Initialize a chain. Logs to {name}.log. Returns a `MiniChain` that
    holds LLM backends..
    """
    return MiniChain(name)


# def show_log(filename: str, o: Callable[[str], Any] = sys.stderr.write) -> None:
#     """
#     Write out the full asynchronous log from file `filename`.
#     """
#     render_tasks(
#         o,
#         tasks_from_iterable([json.loads(line) for line in open(filename)]),
#         colorize=True,
#         human_readable=True,
#     )



import json
from dataclasses import asdict, dataclass, fields, is_dataclass
from enum import Enum
from itertools import count
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from eliot import start_action
from jinja2 import (
    Environment,
    FileSystemLoader,
    PackageLoader,
    Template,
    select_autoescape,
)

from .backend import Backend, MinichainContext, Request

env = Environment(
    loader=PackageLoader("minichain"),
    autoescape=select_autoescape(),
    extensions=["jinja2_highlight.HighlightExtension"],
)


def _prompt(r: Union[str, Request]) -> Request:
    if isinstance(r, str):
        return Request(r)
    else:
        return r


Input = TypeVar("Input")
Output = TypeVar("Output")
FnOutput = TypeVar("FnOutput")


def enum(x: Type[Enum]) -> Dict[str, int]:
    d = {e.name: e.value for e in x}
    return d


def walk(x: Any) -> Any:
    if issubclass(x if get_origin(x) is None else get_origin(x), List):
        return {"_t_": "list", "t": walk(get_args(x)[0])}
    if issubclass(x, Enum):
        return enum(x)

    if is_dataclass(x):
        return {y.name: walk(y.type) for y in fields(x)}
    return x.__name__


def type_to_prompt(out: type) -> str:
    tmp = env.get_template("type_prompt.pmpt.tpl")
    d = walk(out)
    return tmp.render({"typ": d})


def simple(model, **kwargs):  # type: ignore
    return model(kwargs)


@dataclass
class History:
    prompt: "Prompt[Input, Output, FnOutput]"
    inputs: List[Any]


@dataclass
class Fail:
    arg_num: int
    data: Any

    def __str__(self):
        return "fail"


@dataclass
class Chain:
    # TODO: Add caching
    history: History

    def run_gen(self, trial: int = 0, data: Any = None) -> Any:
        args = []
        count = []
        for inp in self.history.inputs:
            inp2 = inp
            if isinstance(inp, Chain):
                for inp2 in inp.run_gen():
                    yield None
            args.append(inp2)
            count.append(0)
        for out in self.history.prompt.expand(args, trial, data):
            yield None
        while isinstance(out, Fail):
            count[out.arg_num] += 1
            inp = self.history.inputs[out.arg_num]
            assert isinstance(inp, Chain)
            for inp2 in inp.run_gen(trial=count[out.arg_num], data=out.data):
                yield None
            args[out.arg_num] = inp2
            for out in self.history.prompt.expand(args, trial, data):
                yield None
            yield None

        yield out

    def run(self) -> Any:
        for x in self.run_gen():
            if x is not None:
                return x


class Prompt(Generic[Input, Output, FnOutput]):
    counter = count()

    def __init__(
        self,
        backend: Backend,
        parser: Union[str, Callable[[str], Output]],
        template_file: Optional[str],
        template: Optional[str],
        stop_template: Optional[str],
        fn: Callable[[Callable[[Input], Output]], FnOutput] = simple,
        stream: bool = False,
        block_input=None,
        block_output=None,
    ):
        self.backend: Backend = backend
        self.stream = stream
        self.description: str = (
            backend.description if hasattr(backend, "description") else ""
        )
        self.parser: Union[str, Callable[[str], Output]] = parser
        self.template_file: Optional[str] = template_file
        self.template: Optional[str] = template
        self.stop_template: Optional[str] = stop_template
        self.block_input = block_input
        self.block_output = block_output
        self.fn = fn
        self._fn: str = fn.__name__
        self._id: int = Prompt.counter.__next__()

    def parse(self, response: str) -> Any:
        """
        Convert from the string response of the function
        to the output type.
        """
        if isinstance(self.parser, str):
            if self.parser == "str":
                return response
            elif self.parser == "json":
                return json.loads(response)
        else:
            return self.parser(response)

    def run_verbose(
        self, r: Union[str, Request], stream=False
    ) -> Tuple[Request, str, Output]:
        with start_action(action_type=str(self.fn)):
            request = _prompt(r)
            with start_action(action_type="Prompted", prompt=request.prompt):
                response: Union[str, Any] = self.backend.run(
                    request
                    if hasattr(self.backend, "needs_request")
                    else request.prompt
                )
            with start_action(action_type="Response", result=response):
                output = self.parse(response)
        if not isinstance(response, str):
            response = "(data)"
        return (request, response, output)

    def run_stream(
        self, r: Union[str, Request], stream=False
    ) -> Tuple[Request, str, Output]:
        request = _prompt(r)
        for r in self.backend.run_stream(
            request if hasattr(self.backend, "needs_request") else request.prompt
        ):
            yield (request, r, r)

    def template_fill(self, inp: Any) -> Request:
        kwargs = inp
        if self.template_file:
            tmp = Environment(loader=FileSystemLoader(".")).get_template(
                name=self.template_file
            )
        elif self.template:
            tmp = Template(self.template)

        x = tmp.render(**kwargs)

        if self.stop_template:
            stop = [Template(self.stop_template).render(**kwargs)]
        else:
            stop = None
        return Request(x, stop)

    def __call__(self, *args: Any) -> FnOutput:
        return Chain(History(self, args))

    class Model:
        def __init__(self, prompt, trial, data):
            self.prompt = prompt
            self.trial = trial
            self.data = data
            self.run_log = None

        def fail(self, argnum: int, data: Any = None) -> Fail:
            return Fail(argnum - 1, data)

        def __call__(self, input_):
            for r in self.stream(input_):
                pass
            return r

        def stream(self, input_):
            assert self.run_log is None, "Only call `model` once per function"
            if (
                self.prompt.template is not None
                or self.prompt.template_file is not None
            ):
                if not isinstance(input_, dict):
                    input_ = asdict(input_)
                input_ = dict(**input_)
                input_["_trial"] = self.trial
                input_["_fail_data"] = self.data

                result = self.prompt.template_fill(input_)
            else:
                result = input_
            if self.prompt.stream and hasattr(self.prompt.backend, "run_stream"):
                self.run_log = ("", "", "")
                for a, b, c in self.prompt.run_stream(result):
                    self.run_log = (a, self.run_log[1] + b, self.run_log[2] + c)
                    yield c
            else:
                self.run_log = (Request(input_), "", "")
                yield None
                self.run_log = self.prompt.run_verbose(result)
                yield self.run_log[-1]

    def expand(self, args, trial=0, data=None):
        MinichainContext.prompt_count.setdefault(self._id, -1)
        if trial == 0:
            MinichainContext.prompt_count[self._id] += 1
        count = MinichainContext.prompt_count[self._id]
        MinichainContext.prompt_store.setdefault((self._id, count), [])
        MinichainContext.prompt_store[self._id, count].append(("", Request(""), "", ""))

        model = self.Model(self, trial, data)

        if self.stream:
            for output in self.fn(model, *args):

                t = model.run_log
                assert model.run_log, str(model)
                run_log = (args, t[0], t[1], output)
                count = MinichainContext.prompt_count[self._id]
                MinichainContext.prompt_store.setdefault((self._id, count), [])
                MinichainContext.prompt_store[self._id, count][-1] = run_log
                yield None
        else:
            output = self.fn(model, *args)

        if not isinstance(output, Fail):
            assert model.run_log, str(model)
            t = model.run_log
            run_log = (args, t[0], t[1], output)
            MinichainContext.prompt_store[self._id, count][-1] = run_log

        yield output


def prompt(
    backend: Backend,
    parser: Union[str, Any] = "str",
    template_file: Optional[str] = None,
    template: Optional[str] = None,
    stop_template: Optional[str] = None,
    stream: bool = False,
    block_input=None,
    block_output=None,
) -> Callable[[Any], Prompt[Input, Output, FnOutput]]:
    return lambda fn: Prompt(
        backend,
        parser,
        template_file,
        template,
        stop_template,
        fn,
        stream,
        block_input,
        block_output,
    )



from dataclasses import dataclass
from jinja2 import (
    Environment,
    FileSystemLoader,
    PackageLoader,
    Template,
    select_autoescape,
)
from typing import Mapping, Any, Optional
from minichain import Prompt, Output, Input, Request

@dataclass
class Template(Prompt[Mapping[str, Any], Output]):
    """
    A prompt that uses Jinja to define a prompt based on a static template.
    Set `template_file` to the Jinja template file.
    """
    template_file: Optional[str] = None
    template: Optional[str] = None
    stop_template: Optional[str] = None
        
    def parse(self, result: str, inp: Mapping[str, Any]) -> Output:
        return str(result)  # type: ignore

    def prompt(self, inp: Input) -> Request:
        kwargs = self.to_dict(inp)
        if self.template_file:
            tmp = Environment(loader=FileSystemLoader(".")).get_template(
                name=self.template_file
            )
        elif self.template:
            tmp = self.template  # type: ignore
        else:
            tmp = Template(self.prompt_template)
        if isinstance(kwargs, dict):
            x = tmp.render(**kwargs)
        else:
            x = tmp.render(**asdict(kwargs))

        if self.stop_template:
            stop = [Template(self.stop_template).render(**kwargs)]
        else:
            stop = None
        return Request(x, stop)



import inspect
import os
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, List, Set, Tuple, Union

import gradio as gr
from gradio.blocks import Block

from minichain import start_chain

from .backend import MinichainContext
from .base import Prompt

CSS = """
#clean div.form {border: 0px}
#response {border: 0px; background: #ffeec6}
#prompt {border: 0px;background: aliceblue}
#json {border: 0px}
span.head {font-size: 60pt; font-family: cursive;}
div.gradio-container {color: black}
div.form {background: inherit}
div.form div {background: inherit}
"""
# #result {border: 0px; background: #c5e0e5}
# #inner {margin: 10px; padding: 10px; font-size: 20px; }
# #inner textarea {border: 0px}

# body {
#   --text-sm: 15px;
#   --text-md: 20px;
#   --text-lg: 22px;
#   --input-text-size: 20px;
#   --section-text-size: 20px;
# }


@dataclass
class HTML:
    html: str

    def _repr_html_(self) -> str:
        return self.html


@dataclass
class DisplayOptions:
    markdown: bool = True


all_data = gr.State({})
final_output = gr.State({})


@dataclass
class Constructor:
    fns: List[Callable[[Dict[Block, Any]], Dict[Block, Any]]] = field(
        default_factory=list
    )
    inputs: Set[Block] = field(default_factory=set)
    outputs: Set[Block] = field(default_factory=set)

    def merge(self, other: "Constructor") -> "Constructor":
        return Constructor(
            self.fns + other.fns,
            self.inputs | other.inputs,
            self.outputs | other.outputs,
        )

    def add_inputs(self, inputs: List[Block]) -> "Constructor":
        return Constructor(self.fns, self.inputs | set(inputs), self.outputs)

    def fn(self, data: Dict[Block, Any]) -> Dict[Block, Any]:
        out: Dict[Block, Any] = {}
        for fn in self.fns:
            out = {**out, **fn(data)}
        return out


def to_gradio_block(
    base_prompt: Prompt[Any, Any, Any],
    i: int,
    display_options: DisplayOptions = DisplayOptions(),
) -> Constructor:

    with gr.Accordion(label=f"👩  Prompt: {str(base_prompt._fn)}", elem_id="prompt"):
        if base_prompt.block_input is not None:
            prompt = base_prompt.block_input()
        elif hasattr(base_prompt.backend, "block_input"):
            prompt = base_prompt.backend.block_input()
        else:
            prompt = gr.Textbox(label="")

    with gr.Accordion(label="💻", elem_id="response"):
        if base_prompt.block_output is not None:
            result = base_prompt.block_output()
        elif hasattr(base_prompt.backend, "block_output"):
            result = base_prompt.backend.block_output()
        else:
            result = gr.Textbox(label="")

    with gr.Accordion(label="...", elem_id="advanced", open=False):
        gr.Markdown(f"Backend: {base_prompt.backend}", elem_id="json")
        input = gr.JSON(elem_id="json", label="Input")
        json = gr.JSON(elem_id="json", label="Output")
        trial = gr.JSON(elem_id="json", label="Previous Trial")

        if base_prompt.template_file:
            # gr.Markdown(f"<center>{base_prompt.template_file}</center>")
            gr.Code(
                label=f"Template: {base_prompt.template_file}",
                value=open(base_prompt.template_file).read(),
                elem_id="inner",
            )
            # btn = gr.Button("Modify Template")
            # if base_prompt.template_file is not None:

            #     def update_template(template: str) -> None:
            #         if base_prompt.template_file is not None:
            #             with open(base_prompt.template_file, "w") as doc:
            #                 doc.write(template)

            #     btn.click(update_template, inputs=c)

    def update(data: Dict[Block, Any]) -> Dict[Block, Any]:
        prev_request_ = ""
        if (base_prompt._id, i) not in data[all_data]:
            return {}

        if (base_prompt._id, i - 1) in data[all_data]:
            prev_request_ = data[all_data][base_prompt._id, i - 1][-1][1].prompt
        trials = len(data[all_data][base_prompt._id, i])
        input_, request_, response_, output_ = data[all_data][base_prompt._id, i][-1]

        def format(s: Any) -> Any:
            if isinstance(s, str):
                return {"string": s}
            return s

        def mark(s: Any) -> Any:
            return str(s)  # f"```text\n{s}\n```"

        j = 0
        for (a, b) in zip(request_.prompt, prev_request_):
            if a != b:
                break
            j += 1

        if j > 30:
            new_prompt = "...\n" + request_.prompt[j:]
        else:
            new_prompt = request_.prompt

        if trials == 1:
            previous_trials = []
        else:
            trial_input_, trial_request_, trial_response_, trial_output_ = data[
                all_data
            ][base_prompt._id, i][trials - 2]
            previous_trials = {
                "input": trial_input_,
                "prompt": trial_request_.prompt,
                "response": trial_response_,
                "output": trial_output_,
            }
        ret = {
            input: format(input_),
            prompt: mark(new_prompt),
            result: mark(output_),
            json: format(response_),
            trial: previous_trials,
        }
        return ret

    return Constructor([update], set(), {input, prompt, result, json, trial})


def chain_blocks(prompts: List[Prompt[Any, Any, Any]]) -> Constructor:
    cons = Constructor()
    count = {}
    for p in prompts:
        count.setdefault(p._id, 0)
        i = count[p._id]
        cons = cons.merge(to_gradio_block(p, i))
        count[p._id] += 1
    return cons


def api_keys(keys: Set[str] = {"OPENAI_API_KEY"}) -> None:
    if all([k in os.environ for k in keys]):
        return
    key_names = {}

    with gr.Accordion(label="API Keys", elem_id="json", open=False):
        if "OPENAI_API_KEY" in keys and "OPENAI_API_KEY" not in os.environ:
            key_names["OPENAI_API_KEY"] = gr.Textbox(
                os.environ.get("OPENAI_API_KEY"),
                label="OpenAI Key",
                elem_id="json",
                type="password",
            )
            gr.Markdown(
                """
            * [OpenAI Key](https://platform.openai.com/account/api-keys)
            """
            )

        if "HF_KEY" in keys:
            gr.Markdown(
                """
            * [Hugging Face Key](https://huggingface.co/settings/tokens)
            """
            )

            key_names["HF_KEY"] = gr.Textbox(
                os.environ.get("HF_KEY"),
                label="Hugging Face Key",
                elem_id="inner",
                type="password",
            )

        if "SERP_KEY" in keys:
            gr.Markdown(
                """
            * [Search Key](https://serpapi.com/users/sign_in)
            """
            )
            key_names["SERP_KEY"] = gr.Textbox(
                os.environ.get("SERP_KEY"),
                label="Search Key",
                elem_id="inner",
                type="password",
            )

        api_btn = gr.Button("Save")

    def api_run(data):  # type: ignore
        for k, v in key_names.items():
            if data[v] is not None and data[v] != "":
                os.environ[k] = data[v]
        return {}

    api_btn.click(api_run, inputs=set(key_names.values()))


def show(
    prompt: Prompt[Any, Any, Any],
    examples: Union[List[str], List[Tuple[str]]] = [""],
    subprompts: List[Prompt[Any, Any, Any]] = [],
    fields: List[str] = [],
    initial_state: Any = None,
    out_type: str = "markdown",
    keys: Set[str] = {"OPENAI_API_KEY"},
    description: str = "",
    code: str = "",
    css="",
) -> gr.Blocks:
    """
    Constructs a gradio component to show a prompt chain.

    Args:
        prompt: A prompt or prompt chain to display.
        examples: A list of example inputs, either string or tuples of fields
        subprompts: The `Prompt` objects to display.
        fields: The names of the field input to the prompt.
        initial_state: For stateful prompts, the initial value.
        out_type: type of final output
        keys: user keys required
        description: description of the model
        code: code to display
        css : additional css

    Returns:
        Gradio block
    """
    fields = [arg for arg in inspect.getfullargspec(prompt).args if arg != "state"]
    with gr.Blocks(css=CSS + "\n" + css, theme=gr.themes.Monochrome()) as demo:
        # API Keys
        api_keys()

        constructor = Constructor()

        # Collect all the inputs
        state = gr.State(initial_state)
        constructor = constructor.merge(Constructor(inputs={state}, outputs={state}))

        # Show the description
        gr.Markdown(description)

        # Build the top query box with one input for each field.
        inputs = list([gr.Textbox(label=f) for f in fields])
        examples = gr.Examples(examples=examples, inputs=inputs)
        query_btn = gr.Button(label="Run")
        constructor = constructor.add_inputs(inputs)

        # Intermediate prompt displays
        constructor = constructor.merge(chain_blocks(subprompts))

        # Final Output result
        with gr.Accordion(label="✔️", elem_id="result"):
            typ = gr.JSON if out_type == "json" else gr.Markdown
            output = typ(elem_id="inner")

        def output_fn(data: Dict[Block, Any]) -> Dict[Block, Any]:
            final = data[final_output]
            return {state: final, output: final}

        constructor = constructor.merge(Constructor([output_fn], set(), {output}))

        def run(data):  # type: ignore
            prompt_inputs = {k: data[v] for k, v in zip(fields, inputs)}
            if initial_state is not None:
                prompt_inputs["state"] = data[state]

            with start_chain("temp"):

                for output in prompt(**prompt_inputs).run_gen():
                    data[all_data] = dict(MinichainContext.prompt_store)
                    data[final_output] = output
                    yield constructor.fn(data)
                    if output is not None:
                        break
            yield constructor.fn(data)

        query_btn.click(run, inputs=constructor.inputs, outputs=constructor.outputs)

        if code:
            gr.Code(code, language="python", elem_id="inner")

    return demo



class Function:
    def __init__(self):
        pass
        
    def render(self, *args) -> str:
        pass

    def parse(self, output: str) -> Any:
        return output

class Backend(Proto):
    async def run(self, inp: str)-> str: ...

class Bot:
    def __init__(self, backend: Backend):
        self.backend = backend

    async def __call__(self, fn) -> Any:
        prompt = fn.render()
        result = await self.backend.run(prompt)
        return fn.parse(result)
    
def composite(fn):
    return lambda *args: async lambda bot: return fn(bot, *args)

def simple(fn):
    class TempFunction(Function):
        def render(self):
            return fn(output)
    return TempFunction


